#define detect 3						//детектор нуля
#define triac 5							//тиристор
#define zerooffset 10					// Сдвиг срабатывания детектора нуля относительно реального нуля в тиках таймера
#define c_timer 625-zerooffset	        // Максимальное количество срабатываний таймера за полупериод синусоиды
#include <ReadDigKey.h>
ReadDigKey key;
#include <Wire.h>
#include <LiquidCrystal_I2C.h> //https://github.com/marcoschwartz/LiquidCrystal_I2C
LiquidCrystal_I2C lcd(0x27, 16, 2);

#define pSW 7			// энкодер SW кнопка
#define pCLK 8          // Pin для энкодера CLK
#define pDT 9			// Pin для энкодера DT

#define ACS_coeff 0.048828125	// Коэффициент датчика ACS712 |5А - 0.024414063 | 20А - 0.048828125 | 30A - 0.073242188 |
#define resist_adres 0									// адрес в eeprom для хранения сопротивления нагрузки
#define log_adres resist_adres + sizeof(uint16_t)			// адрес начала журнала в eeprom
#define quiant_cells 10									// количество ячеек памяти для сохранения мощьности


volatile unsigned int cntr = 0;			// счетчик количества измерений мгновенного значения тока
volatile bool zero = false;				// Начало полупериода
volatile unsigned long Isumm = 0;		// переменная для хранения суммы квадратов
volatile unsigned int tstamp_lcd = 0;	// Количество периодов между выводами в дисплей и serial port

float real_I = 0;		// переменная расчета RMS тока
float inst_I = 0;						// установленный ток
uint16_t coef_inegr;						// Коэффициент интеграции в расчете угла
uint16_t inst_P = 0;					// установливаемая мощность
int buf_inst = 0;

float angle = c_timer;
bool set_menu = false;
bool eprom_menu = false;

bool initialization = false;		// Запускаем начальную инициализацию. Удерживать кнопку энкодера при включении
int ini_u = 220;					// Установка значения напряжения при открытом триаке

uint16_t resist;					// Активное сопротивление нагрузки (ом). Минимум 12 ом
uint16_t max_P = 150;				// Максимальная мощьность нагрузки до при 220 в., 1000 при сопротивлении до 302.5 ом, 2500 до 48.4 ом, 4000 до 19.36
char step_P = 5;					// Шаг установки мощьности нагрузки,  25 при сопротивлении до 302.5 ом, 50 до 48.4 ом, 50 до 19.36


// Объект EEPROM //////////////////////////////////////////
enum _RES { _Prev=1, _Last=0, _Next=-1 };
struct sPwrCell {
	uint16_t Power;
	uint8_t Cell;
};
class Log {
public:
	Log(uint8_t _ferstidx, uint8_t _cells) {
		ferstidx = _ferstidx;
		maxidx = _cells;
		adress = ferstidx + (maxidx * sizeof(int));
		idx = eeprom_read_byte((uint8_t*)adress);
		if (idx == 0xFF) idx = 0;
		last = idx;
	}
	void Move(char _vect) {
		if (_vect == _Last) idx = last;
		if (_vect == _Prev) idx = (--idx >= 0) ? idx : maxidx-1;
		if (_vect == _Next) idx = (++idx < maxidx) ? idx : 0;
	}
	void Add(unsigned int _inst_P) {
		Move(_Last); 
		if (Read().Power != 0) Move(_Next);
		last = idx;
		eeprom_write_word((uint16_t*)(last * sizeof(int) + ferstidx), _inst_P);
		eeprom_write_byte((uint8_t*)adress, last);
	}
	sPwrCell Read() {
		sPwrCell rd;
		rd.Power = eeprom_read_word((uint16_t*)(idx * sizeof(int) + ferstidx));
		if (rd.Power == 0xFFFF) rd.Power = 0;
		rd.Cell= (idx > last) ? 10 - (idx - last) : last - idx;
		return rd;
	}
private:
	int idx;
	uint8_t ferstidx, maxidx, last=0, adress=0;
} elog(log_adres, quiant_cells);

void(*resetFunc) (void) = 0;		// Перезагрузка Ардуины

//======================================================== Прерывание по нулю
void zero_crosss_int() {
	TCNT1 = 0;
	PORTD &= ~(1 << PORTD5);
	zero=true;
	tstamp_lcd++;
	key.readencoder();
}
//======================================================== Обработка таймера по совпадению нуля
ISR(TIMER1_COMPA_vect) {
	if (TCNT1 < c_timer) PORTD |= (1 << PORTD5);
}
//======================================================== Обработка прерывания АЦП для расчета среднеквадратичного тока
ISR(ADC_vect) {
	unsigned long Iism = 0;	//мгновенные значения тока
	byte An_pin = ADCL;
	byte An = ADCH;
	if (cntr < 1024) {
		Iism = ((An << 8) + An_pin) - 512;
		Iism *= Iism;                    // возводим значение в квадрат
		Isumm += Iism;                   // складываем квадраты измерений
		cntr++;
	}
}
void setup(void) {
	Serial.begin(115200);		 //активируем прием сообщения
	pinMode(detect, INPUT);          //детектор нуля
	pinMode(triac, OUTPUT);          //тиристор
	PORTD &= ~(1 << PORTD5);
	key.add_key(pSW);
	key.add_enccoder(pCLK, pDT);
	lcd.init();
	lcd.backlight();
	//===========================================================Настройка АЦП
	ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); //
	ADCSRA = B11101111; //Включение АЦП
	ACSR = (1 << ACD);
	//------ Timer1 ---------- Таймер задержки времени открытия триака после детектирования нуля (0 триак не откроется)
	TCCR1A = 0x00;  //
	TCCR1B = 0x00;    //
	TCCR1B = (1 << CS12) | (0 << CS11) | (0 << CS10); // Тактирование от CLK.
	OCR1A = 0;                   // Верхняя граница счета. Диапазон от 0 до 65535.
	TIMSK1 |= (1 << OCIE1A);     // Разрешить прерывание по совпадению
	attachInterrupt(1, zero_crosss_int, RISING);//вызов прерывания при детектировании нуля
	// Читаем сопротивление нагрузки из EEPROM
	resist = eeprom_read_word((uint16_t*)(resist_adres));
	initialization = !digitalRead(pSW) || (resist == 0xFFFF);
	if (initialization) {
		for (byte i = resist_adres; i <= log_adres + 1 + (quiant_cells * sizeof(int)); i++) {
			eeprom_write_byte((uint8_t*)i, 0xFF);	// очистить eeprom
		}
	} else {
		if (resist >= 301) {
			max_P = 160; step_P = 1; coef_inegr = 16;
		} else if (resist >= 49) {
			max_P = 1000; step_P = 5; coef_inegr = 100;
		} else if (resist >= 19) {
			max_P = 2500; step_P = 10; coef_inegr = 200;
		} else if (resist > 0) {
			max_P = 4000; step_P = 10; coef_inegr = 300;
		}
	}
}

void loop(void) {
	if (zero && cntr == 1024) {
		Isumm >>= 10;
		#define averageFactor 5
		real_I = (real_I * (averageFactor - 1) + ((Isumm>2) ? sqrt(Isumm) * ACS_coeff : 0)) / averageFactor;
		zero = false;
		Isumm = cntr = 0;									// обнуляем суммы токов и счетчики
	}
	if (initialization) {
		angle = zerooffset;
	} else {
		if (inst_I) {										// Расчет угла открытия триака
			angle += (real_I - inst_I)  / coef_inegr;
			angle = constrain(angle, zerooffset, c_timer);
		} else
			angle = c_timer;
	}
	OCR1A = int(angle);
// Обработака кнопки, энкодера и меню
	key.readkey();
	static char cmenu = 0;
	if (key.shot_press()) {			 			// Короткое нажатие кнопки
		cmenu = (set_menu && buf_inst != inst_P) ? 0 : (++cmenu) % 3;
		set_menu = (cmenu == 1);
		eprom_menu = (cmenu == 2);
		elog.Move(_Last);
		buf_inst = (set_menu) ? inst_P : (eprom_menu) ? elog.Read().Power : buf_inst;
	}
	if (key.long_press()) {						// Долгое нажатие кнопки
		if (set_menu || eprom_menu) {
			inst_P = buf_inst;
			inst_I = sqrt(inst_P/float(resist));
			if (set_menu) elog.Add(inst_P);
		} else if (initialization) {					// Первоначальная инициализация
			lcd.setCursor(0, 0); lcd.print("      Save      ");
			lcd.setCursor(0, 1); lcd.print("    parametrs   ");
			eeprom_write_word((uint16_t*)(resist_adres), uint16_t(ini_u / real_I));
			delay(3000);
			resetFunc();
		} else
			inst_P = buf_inst = inst_I = 0;
		cmenu = set_menu = eprom_menu = false;
	}
	// Энкодер
//	key.readencoder();
	if (key.encoder) {									// вращение
		if (initialization) {							// Первоначальная инициализация
			ini_u += int(key.encoder);
			ini_u = ini_u < 180 ? 240 : (ini_u > 240) ? 180 : ini_u;
		} else {
			if (set_menu) {
				buf_inst += int(step_P * key.encoder*abs(key.encoder));
				buf_inst = buf_inst < 0 ? max_P : (buf_inst > max_P) ? 0 : buf_inst;
			} else if (eprom_menu) {
				elog.Move(key.encoder / abs(key.encoder));
				buf_inst = elog.Read().Power;
			}
			set_menu = !eprom_menu;
		}
		key.encoder = 0;
	}
	if (key.previous_millis_down + 100000 < millis()) set_menu = eprom_menu = false;
	//=========================Информация в порт и дисплей
	if (!(tstamp_lcd  % 10)) {
		static char buff0[17] = { "                " };
		static char buff1[17] = { "                " };
		if (initialization) {					// Первоначальная инициализация
			char fbuf[6] = {"     "};
			dtostrf(real_I, 5, 2, fbuf);
			sprintf(buff0, "Elect %sA", fbuf);
			sprintf(buff1, "Volt    %3dV", ini_u);
		} else {
			if (!(tstamp_lcd % 100)) {
				if (real_I>0.2) 
					sprintf(buff0, "Cur %4d", (uint16_t)(pow(real_I, 2)*resist));
				else 
					sprintf(buff0, "Cur ----");
			}
			if (set_menu) 
				sprintf(buff1, "Ins %4d << %4d", inst_P, buf_inst);
			else if (eprom_menu)
				sprintf(buff1, "Ins %4d M%1d %4d", inst_P, elog.Read().Cell, buf_inst);
			else
				sprintf(buff1, "Ins %4d        ", inst_P);
		}
		lcd.setCursor(0, 0); lcd.print(buff0);
		lcd.setCursor(0, 1); lcd.print(buff1);
	}
}







